__author__ = "Richard Mwewa"
__about__ = "https://about.me/rly0nheart"
__version__ = "1.1.1"

__description__ = """# Orbis Unum
> 🌍 IP Geolocator & Coordinate Mapper 📍"""
__epilog__ = f"# by [{__author__}]({__about__})"



from typing import Union


class Messages:
    @staticmethod
    def requesting_data(
        request_type: str, target_title: str, target: Union[str, tuple]
    ) -> str:
        return f"Requesting [green][italic]{request_type}[/][/] data for {target_title}: {target}..."

    @staticmethod
    def loaded_addresses(filename: str) -> str:
        return f"Loaded IP Addresses from file: [green][link file://{filename}]{filename}[/]"

    @staticmethod
    def opening_map(map_name: str) -> str:
        return f"Opening generated map: [link file://{map_name}]{map_name}..."

    @staticmethod
    def an_error_occurred(error: str) -> str:
        return f"An error occurred: [red]{error}[/]"



import argparse
import getpass
import logging
import os
import subprocess
from datetime import datetime
from typing import Optional, Union

import aiohttp
from rich import print
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter

from . import __description__, __epilog__, __version__


async def send_request(endpoint: str) -> Optional[Union[dict, list, None]]:
    """
    Asynchronously sends a GET request to a specified API endpoint and returns the JSON data from it.

    :param endpoint: The API endpoint to send the request to.
    :return: Returns JSON data as a dictionary or list. Returns None if fetching fails.
    """
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(endpoint) as response:
                if response.status == 200:
                    data = await response.json()
                    return data
                else:
                    error_message = await response.json()
                    log.error(f"An api error occurred: {error_message}")
    except aiohttp.ClientError as error:
        log.error(f"An http error occurred: {error}")
    except Exception as error:
        log.critical(f"An unknown error occurred: {error}")


async def check_updates():
    """
    Checks for latest updates by retrieving the release tag from the releases page of the program from GitHub
    Then compares the remote version tag with the tag in the program.
    If they match, program assumes it's up-to-date.
    If not, print a message notifying the user about the remote version (which is treated as the official new release)
    , and lastly prints the release notes of the presumed new release.
    """
    release_data = await send_request(
        "https://api.github.com/repos/rly0nheart/orbis-unum/releases/latest"
    )
    if release_data.get("tag_name"):
        remote_version = release_data.get("tag_name")
        if remote_version != __version__:
            markdown_release_notes = Markdown(release_data.get("body"))
            log.info(
                f"Orbis Unum {remote_version} (from {__version__}) is available. "
                f"To update, run [italic][green]pip install --upgrade orbis-unum[/][/]"
            )
            print(markdown_release_notes)


def path_finder(directories: list) -> None:
    """
    Checks if the specified directories exist.
    If not, it creates them.

    :param directories: List of directories to check and create
    :return: None
    """
    # Construct path to the user's home directory
    home_directory = os.path.expanduser("~")

    for directory in directories:
        # Construct and create each directory from the directories list if it doesn't already exist
        os.makedirs(os.path.join(home_directory, directory), exist_ok=True)


def clear_screen():  # -> A cleared screen
    """
    Clear the terminal screen/
    If Operating system is Windows, uses the 'cls' command. Otherwise, uses the 'clear' command

    :return: Uhh, a cleared screen? haha
    """
    subprocess.call("cmd.exe /c cls" if os.name == "nt" else "clear")


def get_current_user() -> str:
    """
    Gets current logged-in user's username.

    :return: Current user's username.
    """
    return getpass.getuser()


def format_map_name(map_name: str) -> str:
    """
    This function takes a string and adds a timestamp to it in a different format depending on the operating system.

    :param map_name: A string to which the timestamp will be appended.
    :return: A string with the timestamp appended.
    """

    # Check the name of the operating system using os.name
    if os.name == "nt":
        # If the operating system is Windows, then use this format for the timestamp
        formatted_map_name = datetime.now().strftime(f"{map_name}_%d-%m-%Y %H-%M-%S%p")
    else:
        # If the operating system is not Windows (e.g., Unix or Linux), then use this format for the timestamp
        formatted_map_name = datetime.now().strftime(f"{map_name}_%d-%m-%Y %H:%M:%S%p")

    # Return the formatted map name
    return formatted_map_name


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=Markdown(__description__, style="argparse.text"),
        epilog=Markdown(__epilog__, style="argparse.text"),
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument("-i", "--ip", help="ip address or file containing ip addresses")
    parser.add_argument(
        "-o", "--output", help="map output name (default %(default)s)", default="orbis"
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="run orbis-cli in debug mode (also applies to orbis-web)",
        action="store_true",
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    return parser


def set_loglevel(debug_mode: bool) -> logging.getLogger:
    """
    Configure and return a logging object with the specified log level.

    :param debug_mode: If True, the log level is set to "NOTSET". Otherwise, it is set to "INFO".zzz
    :return: A logging object configured with the specified log level.
    """
    logging.basicConfig(
        level="NOTSET" if debug_mode else "INFO",
        format="%(message)s",
        handlers=[
            RichHandler(markup=True, log_time_format="%H:%M:%S", show_level=True)
        ],
    )
    return logging.getLogger("Orbis Unum")


args = create_parser().parse_args()
log = set_loglevel(debug_mode=args.debug)

# Construct the path to the directory containing the current script
CURRENT_FILE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

# Construct the path to the templates directory
TEMPLATES_DIRECTORY = os.path.join(CURRENT_FILE_DIRECTORY, "templates")



import webbrowser

from .coreutils import args, log, path_finder, check_updates, Markdown, print
from .orbis import message, app, create_map, get_location_data_list


def orbis():
    print(
        Markdown(
            markup="""
# Orbis Unum

## CLI
> Use the `orbis-cli` command to start the command-line interface variant of **Orbis Unum**. 
In CLI mode, you should provide an IP Address or a file containing IP Addresses as an argument. 
Orbis will fetch the geolocation information for the given IP Address(es) and display it on a map.
    
```
orbis-cli --ip 8.8.8.8/orbis-cli ips.txt
```


## WEB
> Use the `orbis-web` command to start the web variant of **Orbis Unum**.
In WEB mode, you will be presented with a form where you can input geographical coordinates 
(latitude and longitude). After submitting the form, Orbis will fetch the geolocation information for the 
given coordinates and display it on a map.
    
```
orbis-web
```
    
### Note
> The web interface will be accessible at http://localhost:5000/index after you run the 'orbis-web' command.
"""
        )
    )


def orbis_cli():
    """
    The CLI variant of Orbis Unum.
    """
    import asyncio

    if args.ip:
        # Create the directory where generated maps will be stored.
        path_finder(["orbis-maps"])

        asyncio.run(check_updates())
        map_data = asyncio.run(get_location_data_list(ip_address=args.ip))
        created_map = create_map(map_data=map_data, map_name=args.output)

        log.info(message.opening_map(created_map))
        webbrowser.open(created_map)
    else:
        log.info("use orbis-cli -h/--help to show usage.")


def orbis_web():
    """
    The web variant of Orbis Unum.
    """
    check_updates()

    # Start the Flask app
    log.info("Hosting Flask app at http://localhost:5000/index")
    app.run(debug=args.debug)



import asyncio
import os
from datetime import datetime

from flask import Flask, request, redirect, url_for, render_template
from jinja2 import Environment, FileSystemLoader
from markupsafe import Markup

from .coreutils import (
    log,
    send_request,
    get_current_user,
    format_map_name,
    CURRENT_FILE_DIRECTORY,
    TEMPLATES_DIRECTORY,
)
from .messages import Messages

# Initialise messages
message = Messages()

# Create the Flask application
app = Flask("Orbis Unum: Web Instance", template_folder=TEMPLATES_DIRECTORY)

# List to store location data
location_data = []


@app.route("/index", methods=["GET", "POST"])
def input_coordinates():
    """
    The route for inputting coordinates in the web variant of Orbis.
    It either renders a form for inputting coordinates or processes
    the coordinates that have been inputted.

    :return: Renders the form for a GET request or redirects to the map view for a POST request.
    """
    error = None
    if request.method == "POST":
        try:
            # Get the string of coordinates from the form
            coordinates_input = request.form.get("coordinates")

            # Split the string into a list of tuples, where each tuple is a pair of coordinates
            coordinates_list = [
                tuple(map(float, coordinates.split(",")))
                for coordinates in coordinates_input.split("\n")
            ]

            # Get the location data for each pair of coordinates
            # and create a list of dictionaries where each dictionary has a 'location'
            # key (with the location data) and an 'ip_address' key (with a None value as IP address is not available)
            location_data[:] = [
                {
                    "location": asyncio.run(get_reverse_geolocation_data(coordinates)),
                    "ip_address": None,
                }
                for coordinates in coordinates_list
            ]

            # Redirect to the map view
            return redirect(url_for("map_view"))

        except Exception as e:
            # Create an error message
            error = Markup(
                message.an_error_occurred(error=str(e))
            )  # mark the error as safe
            log.error(message.an_error_occurred(error=str(e)))

    # Render the input form for a GET request
    return render_template(
        "index.html",
        current_user=get_current_user(),
        current_year=datetime.now().year,
        error=error,
    )


@app.route("/map")
def map_view(map_name: str = "Orbis Web"):
    """
    The route for displaying the map in the web variant of Orbis. It uses the map.html template.
    """
    return render_template(
        "map.html",
        map_data=location_data,
        current_user=get_current_user(),
        map_name=map_name,
    )


def process_user_input(user_input: str) -> list:
    """
    Processes input from user to determine the type, and how to return the results.

    :param user_input: IP; could be a single IP or a text file containing IP Addresses.
    :return: A list containing an IP Address/Addresses.
    """
    if os.path.isfile(user_input):
        with open(user_input, "r") as file:
            log.info(message.loaded_addresses(filename=file.name))
            ip_addresses = [ip.strip() for ip in file.readlines()]
            return ip_addresses
    else:
        return [user_input]


async def get_reverse_geolocation_data(coordinates: tuple) -> dict:
    """
    Gets location data from the given coordinates.

    :param coordinates: A tuple containing the longitude and latitude parts of the coordinates.
    :return: A dictionary of the coordinates' location data.
    """
    log.info(
        message.requesting_data(
            request_type="reverse-geolocation",
            target_title="coordinates",
            target=coordinates,
        )
    )
    coordinates_location_data = await send_request(
        f"https://nominatim.openstreetmap.org/reverse?lat={coordinates[0]}&"
        f"lon={coordinates[1]}&format=json&addressdetails=1"
    )

    return coordinates_location_data


async def get_geolocation_data(ip_address: str) -> tuple:
    """
    Gets geolocation data from the given ip address.

    :param ip_address: Target IP Address.
    :return: A tuple containing ip address, latitude and longitude.
    """

    log.info(
        message.requesting_data(
            request_type="geolocation", target_title="IP", target=ip_address
        )
    )
    ip_coordinates = await send_request(f"http://ip-api.com/json/{ip_address}")
    return (
        ip_coordinates.get("query"),
        ip_coordinates.get("lat"),
        ip_coordinates.get("lon"),
    )


async def get_location_data_list(ip_address: str) -> list:
    """
    Given an IP address, this function returns a list of location data, the IP address itself, and the corresponding
    first coordinate for each IP address.

    :param ip_address: A string containing one or more IP addresses (in a file).
    :return: A list of dictionaries where each dictionary contains the
     location data of an IP address and the IP address itself.
    """
    data_list = []
    ip_addresses = process_user_input(user_input=ip_address)
    coordinates_list = [await get_geolocation_data(ip) for ip in ip_addresses]

    for coordinates in coordinates_list:
        data = {
            "location": await get_reverse_geolocation_data(
                (coordinates[1], coordinates[2])
            ),
            "ip_address": coordinates[0],
        }
        data_list.append(data)

    return data_list


def create_map(map_data: list, map_name: str) -> str:
    """
    Uses the map template to create a new map with the geolocation data returned from the get_location_data function.

    :param map_data: List of dictionaries containing the location data of each IP Address and the IP itself.
    :param map_name: Output filename of the generated map.
    :return: An interactive map in default browser
    (with pins pointing on the areas that correspond the IPs coordinates).
    """
    home_directory = os.path.expanduser("~")
    maps_directory = os.path.join(home_directory, "orbis-maps")

    env = Environment(
        loader=FileSystemLoader(os.path.join(CURRENT_FILE_DIRECTORY, "templates"))
    )

    map_template = env.get_template("map.html")

    rendered_map = map_template.render(
        current_user=get_current_user(),
        current_datetime=datetime.now(),
        map_name=map_name,
        map_data=map_data,
    )

    with open(
        os.path.join(maps_directory, f"{format_map_name(map_name)}.html"),
        "w",
        encoding="utf-8",
    ) as created_map:
        created_map.write(rendered_map)

    return created_map.name
